@@ -78,7 +78,7 @@ QT_BEGIN_NAMESPACE
 bool QSerialPortPrivate::open(QIODevice::OpenMode mode)
 {
     DWORD desiredAccess = 0;
-    originalEventMask = EV_ERR;
+    originalEventMask = 0;
 
     if (mode & QIODevice::ReadOnly) {
         desiredAccess |= GENERIC_READ;
@@ -104,36 +104,26 @@ bool QSerialPortPrivate::open(QIODevice::OpenMode mode)
 
 void QSerialPortPrivate::close()
 {
-    if (!::CancelIo(handle))
-        setError(getSystemError());
+    ::CancelIo(handle);
 
-    if (notifier) {
-        delete notifier;
-        notifier = Q_NULLPTR;
-    }
+    delete notifier;
+    notifier = Q_NULLPTR;
 
-    if (startAsyncWriteTimer) {
-        delete startAsyncWriteTimer;
-        startAsyncWriteTimer = Q_NULLPTR;
-    }
+    delete startAsyncWriteTimer;
+    startAsyncWriteTimer = Q_NULLPTR;
 
+    communicationStarted = false;
     readStarted = false;
     writeStarted = false;
     writeBuffer.clear();
     actualBytesToWrite = 0;
 
-    parityErrorOccurred = false;
-
     if (settingsRestoredOnClose) {
-        if (!::SetCommState(handle, &restoredDcb))
-            setError(getSystemError());
-        else if (!::SetCommTimeouts(handle, &restoredCommTimeouts))
-            setError(getSystemError());
+        ::SetCommState(handle, &restoredDcb);
+        ::SetCommTimeouts(handle, &restoredCommTimeouts);
     }
 
-    if (!::CloseHandle(handle))
-        setError(getSystemError());
-
+    ::CloseHandle(handle);
     handle = INVALID_HANDLE_VALUE;
 }
 
@@ -180,8 +170,12 @@ bool QSerialPortPrivate::setDataTerminalReady(bool set)
         return false;
     }
 
-    currentDcb.fDtrControl = set ? DTR_CONTROL_ENABLE : DTR_CONTROL_DISABLE;
-    return true;
+    DCB dcb;
+    if (!getDcb(&dcb))
+        return false;
+
+    dcb.fDtrControl = set ? DTR_CONTROL_ENABLE : DTR_CONTROL_DISABLE;
+    return setDcb(&dcb);
 }
 
 bool QSerialPortPrivate::setRequestToSend(bool set)
@@ -257,7 +251,8 @@ bool QSerialPortPrivate::waitForReadyRead(int msecs)
     stopWatch.start();
 
     do {
-        OVERLAPPED *overlapped = waitForNotified(timeoutValue(msecs, stopWatch.elapsed()));
+        OVERLAPPED *overlapped = waitForNotified(
+                    qt_subtract_from_timeout(msecs, stopWatch.elapsed()));
         if (!overlapped)
             return false;
 
@@ -273,7 +268,7 @@ bool QSerialPortPrivate::waitForReadyRead(int msecs)
             }
         }
 
-    } while (msecs == -1 || timeoutValue(msecs, stopWatch.elapsed()) > 0);
+    } while (msecs == -1 || qt_subtract_from_timeout(msecs, stopWatch.elapsed()) > 0);
 
     return false;
 }
@@ -290,7 +285,8 @@ bool QSerialPortPrivate::waitForBytesWritten(int msecs)
     stopWatch.start();
 
     forever {
-        OVERLAPPED *overlapped = waitForNotified(timeoutValue(msecs, stopWatch.elapsed()));
+        OVERLAPPED *overlapped = waitForNotified(
+                    qt_subtract_from_timeout(msecs, stopWatch.elapsed()));
         if (!overlapped)
             return false;
 
@@ -312,98 +308,113 @@ bool QSerialPortPrivate::setBaudRate(qint32 baudRate, QSerialPort::Directions di
         setError(QSerialPortErrorInfo(QSerialPort::UnsupportedOperationError, QSerialPort::tr("Custom baud rate direction is unsupported")));
         return false;
     }
-    currentDcb.BaudRate = baudRate;
-    return updateDcb();
+
+    DCB dcb;
+    if (!getDcb(&dcb))
+        return false;
+
+    dcb.BaudRate = baudRate;
+    return setDcb(&dcb);
 }
 
 bool QSerialPortPrivate::setDataBits(QSerialPort::DataBits dataBits)
 {
-    currentDcb.ByteSize = dataBits;
-    return updateDcb();
+    DCB dcb;
+    if (!getDcb(&dcb))
+        return false;
+
+    dcb.ByteSize = dataBits;
+    return setDcb(&dcb);
 }
 
 bool QSerialPortPrivate::setParity(QSerialPort::Parity parity)
 {
-    currentDcb.fParity = TRUE;
+    DCB dcb;
+    if (!getDcb(&dcb))
+        return false;
+
+    dcb.fParity = TRUE;
     switch (parity) {
     case QSerialPort::NoParity:
-        currentDcb.Parity = NOPARITY;
-        currentDcb.fParity = FALSE;
+        dcb.Parity = NOPARITY;
+        dcb.fParity = FALSE;
         break;
     case QSerialPort::OddParity:
-        currentDcb.Parity = ODDPARITY;
+        dcb.Parity = ODDPARITY;
         break;
     case QSerialPort::EvenParity:
-        currentDcb.Parity = EVENPARITY;
+        dcb.Parity = EVENPARITY;
         break;
     case QSerialPort::MarkParity:
-        currentDcb.Parity = MARKPARITY;
+        dcb.Parity = MARKPARITY;
         break;
     case QSerialPort::SpaceParity:
-        currentDcb.Parity = SPACEPARITY;
+        dcb.Parity = SPACEPARITY;
         break;
     default:
-        currentDcb.Parity = NOPARITY;
-        currentDcb.fParity = FALSE;
+        dcb.Parity = NOPARITY;
+        dcb.fParity = FALSE;
         break;
     }
-    return updateDcb();
+    return setDcb(&dcb);
 }
 
 bool QSerialPortPrivate::setStopBits(QSerialPort::StopBits stopBits)
 {
+    DCB dcb;
+    if (!getDcb(&dcb))
+        return false;
+
     switch (stopBits) {
     case QSerialPort::OneStop:
-        currentDcb.StopBits = ONESTOPBIT;
+        dcb.StopBits = ONESTOPBIT;
         break;
     case QSerialPort::OneAndHalfStop:
-        currentDcb.StopBits = ONE5STOPBITS;
+        dcb.StopBits = ONE5STOPBITS;
         break;
     case QSerialPort::TwoStop:
-        currentDcb.StopBits = TWOSTOPBITS;
+        dcb.StopBits = TWOSTOPBITS;
         break;
     default:
-        currentDcb.StopBits = ONESTOPBIT;
+        dcb.StopBits = ONESTOPBIT;
         break;
     }
-    return updateDcb();
+    return setDcb(&dcb);
 }
 
 bool QSerialPortPrivate::setFlowControl(QSerialPort::FlowControl flowControl)
 {
-    currentDcb.fInX = FALSE;
-    currentDcb.fOutX = FALSE;
-    currentDcb.fOutxCtsFlow = FALSE;
-    currentDcb.fRtsControl = RTS_CONTROL_DISABLE;
+    DCB dcb;
+    if (!getDcb(&dcb))
+        return false;
+
+    dcb.fInX = FALSE;
+    dcb.fOutX = FALSE;
+    dcb.fOutxCtsFlow = FALSE;
+    dcb.fRtsControl = RTS_CONTROL_DISABLE;
     switch (flowControl) {
     case QSerialPort::NoFlowControl:
         break;
     case QSerialPort::SoftwareControl:
-        currentDcb.fInX = TRUE;
-        currentDcb.fOutX = TRUE;
+        dcb.fInX = TRUE;
+        dcb.fOutX = TRUE;
         break;
     case QSerialPort::HardwareControl:
-        currentDcb.fOutxCtsFlow = TRUE;
-        currentDcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
+        dcb.fOutxCtsFlow = TRUE;
+        dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
         break;
     default:
         break;
     }
-    return updateDcb();
-}
-
-bool QSerialPortPrivate::setDataErrorPolicy(QSerialPort::DataErrorPolicy policy)
-{
-    policy = policy;
-    return true;
+    return setDcb(&dcb);
 }
 
 bool QSerialPortPrivate::completeAsyncCommunication(qint64 bytesTransferred)
 {
+    communicationStarted = false;
+
     if (bytesTransferred == qint64(-1))
         return false;
-    if (EV_ERR & triggeredEventMask)
-        handleLineStatusErrors();
 
     return startAsyncRead();
 }
@@ -422,12 +433,15 @@ bool QSerialPortPrivate::completeAsyncRead(qint64 bytesTransferred)
     readStarted = false;
 
     bool result = true;
-    if ((bytesTransferred == ReadChunkSize) && (policy == QSerialPort::IgnorePolicy))
+    if (bytesTransferred == ReadChunkSize
+            || queuedBytesCount(QSerialPort::Input) > 0) {
         result = startAsyncRead();
-    else if (readBufferMaxSize == 0 || readBufferMaxSize > buffer.size())
+    } else if (readBufferMaxSize == 0
+               || readBufferMaxSize > buffer.size()) {
         result = startAsyncCommunication();
+    }
 
-    if ((bytesTransferred > 0) && !emulateErrorPolicy())
+    if (bytesTransferred > 0)
         emitReadyRead();
 
     return result;
@@ -453,6 +467,9 @@ bool QSerialPortPrivate::completeAsyncWrite(qint64 bytesTransferred)
 
 bool QSerialPortPrivate::startAsyncCommunication()
 {
+    if (communicationStarted)
+        return true;
+
     ::ZeroMemory(&communicationOverlapped, sizeof(communicationOverlapped));
     if (!::WaitCommEvent(handle, &triggeredEventMask, &communicationOverlapped)) {
         QSerialPortErrorInfo error = getSystemError();
@@ -463,6 +480,7 @@ bool QSerialPortPrivate::startAsyncCommunication()
             return false;
         }
     }
+    communicationStarted = true;
     return true;
 }
 
@@ -471,7 +489,7 @@ bool QSerialPortPrivate::startAsyncRead()
     if (readStarted)
         return true;
 
-    DWORD bytesToRead = policy == QSerialPort::IgnorePolicy ? ReadChunkSize : 1;
+    DWORD bytesToRead = ReadChunkSize;
 
     if (readBufferMaxSize && bytesToRead > (readBufferMaxSize - buffer.size())) {
         bytesToRead = readBufferMaxSize - buffer.size();
@@ -544,34 +562,6 @@ void QSerialPortPrivate::_q_notified(DWORD numberOfBytes, DWORD errorCode, OVERL
         Q_ASSERT(!"Unknown OVERLAPPED activated");
 }
 
-bool QSerialPortPrivate::emulateErrorPolicy()
-{
-    if (!parityErrorOccurred)
-        return false;
-
-    parityErrorOccurred = false;
-
-    switch (policy) {
-    case QSerialPort::SkipPolicy:
-        buffer.getChar();
-        break;
-    case QSerialPort::PassZeroPolicy:
-        buffer.getChar();
-        buffer.ungetChar('\0');
-        emitReadyRead();
-        break;
-    case QSerialPort::IgnorePolicy:
-        return false;
-    case QSerialPort::StopReceivingPolicy:
-        emitReadyRead();
-        break;
-    default:
-        return false;
-    }
-
-    return true;
-}
-
 void QSerialPortPrivate::emitReadyRead()
 {
     Q_Q(QSerialPort);
@@ -589,7 +579,7 @@ qint64 QSerialPortPrivate::writeData(const char *data, qint64 maxSize)
     if (!writeBuffer.isEmpty() && !writeStarted) {
         if (!startAsyncWriteTimer) {
             startAsyncWriteTimer = new QTimer(q);
-            q->connect(startAsyncWriteTimer, SIGNAL(timeout()), q, SLOT(_q_startAsyncWrite()));
+            QObjectPrivate::connect(startAsyncWriteTimer, &QTimer::timeout, this, &QSerialPortPrivate::_q_startAsyncWrite);
             startAsyncWriteTimer->setSingleShot(true);
         }
         startAsyncWriteTimer->start(0);
@@ -597,26 +587,6 @@ qint64 QSerialPortPrivate::writeData(const char *data, qint64 maxSize)
     return maxSize;
 }
 
-void QSerialPortPrivate::handleLineStatusErrors()
-{
-    DWORD errors = 0;
-    if (!::ClearCommError(handle, &errors, Q_NULLPTR)) {
-        setError(getSystemError());
-        return;
-    }
-
-    if (errors & CE_FRAME) {
-        setError(QSerialPortErrorInfo(QSerialPort::FramingError, QSerialPort::tr("Framing error detected while reading")));
-    } else if (errors & CE_RXPARITY) {
-        setError(QSerialPortErrorInfo(QSerialPort::ParityError, QSerialPort::tr("ParityError error detected while reading")));
-        parityErrorOccurred = true;
-    } else if (errors & CE_BREAK) {
-        setError(QSerialPortErrorInfo(QSerialPort::BreakConditionError, QSerialPort::tr("Break condition detected while reading")));
-    } else {
-        setError(QSerialPortErrorInfo(QSerialPort::UnknownError, QSerialPort::tr("Unknown streaming error")));
-    }
-}
-
 OVERLAPPED *QSerialPortPrivate::waitForNotified(int msecs)
 {
     OVERLAPPED *overlapped = notifier->waitForAnyNotified(msecs);
@@ -627,32 +597,39 @@ OVERLAPPED *QSerialPortPrivate::waitForNotified(int msecs)
     return overlapped;
 }
 
+qint64 QSerialPortPrivate::queuedBytesCount(QSerialPort::Direction direction) const
+{
+    COMSTAT comstat;
+    if (::ClearCommError(handle, Q_NULLPTR, &comstat) == 0)
+        return -1;
+    return (direction == QSerialPort::Input)
+            ? comstat.cbInQue
+            : ((direction == QSerialPort::Output) ? comstat.cbOutQue : -1);
+}
+
 inline bool QSerialPortPrivate::initialize()
 {
     Q_Q(QSerialPort);
 
-    ::ZeroMemory(&restoredDcb, sizeof(restoredDcb));
-    restoredDcb.DCBlength = sizeof(restoredDcb);
-
-    if (!::GetCommState(handle, &restoredDcb)) {
-        setError(getSystemError());
+    DCB dcb;
+    if (!getDcb(&dcb))
         return false;
-    }
 
-    currentDcb = restoredDcb;
-    currentDcb.fBinary = TRUE;
-    currentDcb.fInX = FALSE;
-    currentDcb.fOutX = FALSE;
-    currentDcb.fAbortOnError = FALSE;
-    currentDcb.fNull = FALSE;
-    currentDcb.fErrorChar = FALSE;
+    restoredDcb = dcb;
 
-    if (currentDcb.fDtrControl ==  DTR_CONTROL_HANDSHAKE)
-        currentDcb.fDtrControl = DTR_CONTROL_DISABLE;
+    dcb.fBinary = TRUE;
+    dcb.fInX = FALSE;
+    dcb.fOutX = FALSE;
+    dcb.fAbortOnError = FALSE;
+    dcb.fNull = FALSE;
+    dcb.fErrorChar = FALSE;
 
-    currentDcb.BaudRate = inputBaudRate;
+    if (dcb.fDtrControl == DTR_CONTROL_HANDSHAKE)
+        dcb.fDtrControl = DTR_CONTROL_DISABLE;
 
-    if (!updateDcb())
+    dcb.BaudRate = inputBaudRate;
+
+    if (!setDcb(&dcb))
         return false;
 
     if (!::GetCommTimeouts(handle, &restoredCommTimeouts)) {
@@ -663,8 +640,10 @@ inline bool QSerialPortPrivate::initialize()
     ::ZeroMemory(&currentCommTimeouts, sizeof(currentCommTimeouts));
     currentCommTimeouts.ReadIntervalTimeout = MAXDWORD;
 
-    if (!updateCommTimeouts())
+    if (!::SetCommTimeouts(handle, &currentCommTimeouts)) {
+        setError(getSystemError());
         return false;
+    }
 
     if (!::SetCommMask(handle, originalEventMask)) {
         setError(getSystemError());
@@ -672,29 +651,32 @@ inline bool QSerialPortPrivate::initialize()
     }
 
     notifier = new QWinOverlappedIoNotifier(q);
-    q->connect(notifier, SIGNAL(notified(quint32, quint32, OVERLAPPED*)),
-               q, SLOT(_q_notified(quint32, quint32, OVERLAPPED*)));
+    QObjectPrivate::connect(notifier, &QWinOverlappedIoNotifier::notified,
+               this, &QSerialPortPrivate::_q_notified);
     notifier->setHandle(handle);
     notifier->setEnabled(true);
 
-    if (!startAsyncCommunication())
+    if ((originalEventMask & EV_RXCHAR) && !startAsyncCommunication())
         return false;
 
     return true;
 }
 
-bool QSerialPortPrivate::updateDcb()
+bool QSerialPortPrivate::setDcb(DCB *dcb)
 {
-    if (!::SetCommState(handle, &currentDcb)) {
+    if (!::SetCommState(handle, dcb)) {
         setError(getSystemError());
         return false;
     }
     return true;
 }
 
-bool QSerialPortPrivate::updateCommTimeouts()
+bool QSerialPortPrivate::getDcb(DCB *dcb)
 {
-    if (!::SetCommTimeouts(handle, &currentCommTimeouts)) {
+    ::ZeroMemory(dcb, sizeof(DCB));
+    dcb->DCBlength = sizeof(DCB);
+
+    if (!::GetCommState(handle, dcb)) {
         setError(getSystemError());
         return false;
     }
@@ -722,6 +704,9 @@ QSerialPortErrorInfo QSerialPortPrivate::getSystemError(int systemErrorCode) con
     case ERROR_FILE_NOT_FOUND:
         error.errorCode = QSerialPort::DeviceNotFoundError;
         break;
+    case ERROR_PATH_NOT_FOUND:
+        error.errorCode = QSerialPort::DeviceNotFoundError;
+        break;
     case ERROR_INVALID_NAME:
         error.errorCode = QSerialPort::DeviceNotFoundError;
         break;
@@ -825,24 +810,6 @@ static const QList<qint32> standardBaudRatePairList()
     return standardBaudRatesTable;
 };
 
-qint32 QSerialPortPrivate::baudRateFromSetting(qint32 setting)
-{
-    const QList<qint32> baudRatePairs = standardBaudRatePairList();
-    const QList<qint32>::const_iterator baudRatePairListConstIterator
-            = std::find(baudRatePairs.constBegin(), baudRatePairs.constEnd(), setting);
-
-    return (baudRatePairListConstIterator != baudRatePairs.constEnd()) ? *baudRatePairListConstIterator : 0;
-}
-
-qint32 QSerialPortPrivate::settingFromBaudRate(qint32 baudRate)
-{
-    const QList<qint32> baudRatePairList = standardBaudRatePairList();
-    const QList<qint32>::const_iterator baudRatePairListConstIterator
-            = std::find(baudRatePairList.constBegin(), baudRatePairList.constEnd(), baudRate);
-
-    return (baudRatePairListConstIterator != baudRatePairList.constEnd()) ? *baudRatePairListConstIterator : 0;
-}
-
 QList<qint32> QSerialPortPrivate::standardBaudRates()
 {
     return standardBaudRatePairList();
